
/*******************************************************************************
r.suntrack: This program was written in 2016 by Thomas Huld, based on the r.sun raster module. It works
in a similar way to r.sun but can calculate solar irradiance or irradiation on various sun-tracking surfaces.
(C) 2016 Copyright Thomas Huld, European Commission Joint Research Centre
e-mail: thomas.huld@jrc.ec.europa.eu
*******************************************************************************/
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA.
 */

/*v. 2.0 July 2002, NULL data handling, JH */
/*v. 2.1 January 2003, code optimization by Thomas Huld, JH */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <grass/gis.h>
#include <grass/raster.h>
#include <grass/gprojects.h>
#include "sunradstruct.h"
#include "local_proto.h"
#include "rsunglobals.h"



#define HOURANGLE M_PI/12.
#define NUM_PARTITIONS "10"
#define UNDEF    0.             /* undefined value for terrain aspect */
#define UNDEFZ   -9999.         /* internal undefined value for NULL */
#define SKIP    "1"
#define BIG      1.e20
#define IBIG     32767
#define EPS      1.e-4
#define LINKE    "3.0"
#define SLOPE    "0.01"
#define ASPECT   "270"
#define ALB      "0.2"
#define STEP     "0.5"
#define BSKY      1.0
#define DSKY      1.0
#define DIST     "1.0"

#define SCALING_FACTOR 150.
const double invScale = 1. / SCALING_FACTOR;


#define AMAX1(arg1, arg2) ((arg1) >= (arg2) ? (arg1) : (arg2))
#define AMIN1(arg1, arg2) ((arg1) <= (arg2) ? (arg1) : (arg2))
#define DISTANCE1(x1, x2, y1, y2) (sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2)))

int num_temperatures;

char *str_step;

int numTemperatures()
{
    return num_temperatures;
}

void setNumTemperatures(int val)
{
    num_temperatures = val;
}


const double deg2rad = M_PI / 180.;

const double pihalf = 0.5 * M_PI;

const double pi2 = 2. * M_PI;

const double rad2deg = 180. / M_PI;

char *elevin;

char *aspin;

char *slopein;

char *civiltime = NULL;

char *linkein = NULL;

char *albedo = NULL;

char *numaxestracking = NULL;

char *latin = NULL;

char *coefbh = NULL;

char *coefdh = NULL;

char *longin = NULL;

char *horizon = NULL;

char *beam_rad = NULL;

char *diff_rad = NULL;

char *refl_rad = NULL;

char *glob_pow = NULL;

char **coeftemp = NULL;

char *mapset = NULL;

char *per;

char *shade;

char mapname[1024];

char *modelparameters = NULL;

struct Cell_head cellhd;

struct pj_info iproj;

struct pj_info oproj;

struct History hist;

double joules2(struct SunGeometryConstDay *sunGeom,
               struct SunGeometryVarDay *sunVarGeom,
               struct SunGeometryVarSlope *sunSlopeGeom,
               struct SolarRadVar *sunRadVar,
               struct GridGeometry *gridGeom, float *temperatureData,
               unsigned char *horizonpointer,
               double latitude, double longitude, double sinAxisSlope,
               double cosAxisSlope, BeamRadFunc bRadFunc,
               DiffRadFunc dRadFunc);

int initEfficiencyCoeffs(char *filename);

int INPUT_part(int offset);

int OUTGR(void);

int min(int, int);

int max(int, int);

double powerDeficit(double temp, double rad);

double temperature(double *coeffs, double time);

void cube(int, int);

void (*func) (int, int);

void calculate(int angleloss, double singleSlope, double singleAspect,
               double singleAlbedo, struct GridGeometry gridGeom);
double com_declin(int);


int useTemperatureCorr = 0;

int numAxisTracking = 0;

int typ, n, m, ip, jp;

int d, day;

int saveMemory, numPartitions = 1;

int shadowoffset = 0;

int varCount_global = 0;

int bitCount_global = 0;

int arrayNumInt;

int highIrr;

float **z = NULL, **o = NULL, **s = NULL, **li = NULL, **a = NULL, **la =
    NULL, **longitArray, **cbhr = NULL, **cdhr = NULL, *tempdata;
double op, dp;

double invstepx, invstepy;

double sr_min = 24., sr_max = 0., ss_min = 24., ss_max = 0.;

float **beam, **diff, **refl, **globrad;

unsigned char *horizonarray = NULL;

double civilTime;

double timeOffset = 0., totOffsetTime, longitTime = 0.;

double xmin, xmax, ymin, ymax, zmax = 0.;

double zmult = 1.0;

double declin, step, dist;

double li_max = 0., li_min = 100., al_max = 0., al_min = 1.0, la_max = -90.,
    la_min = 90.;
double offsetx = 0.5, offsety = 0.5;

char *tt, *lt;

double o_orig, z1;

double lum_p;

double horizonStep;

double ltime, tim, timo;

double declination;             /* Contains the negative of the declination at the chosen day */

double beam_e, diff_e, refl_e, bh, dh, rr;

double cbh, cdh;

double TOLER;

double tcoeffs[4];

void cube(int jmin, int imin)
{
}


double temperature(double *coeffs, double time)
{
    return coeffs[0] * time * time * time + coeffs[1] * time * time +
        coeffs[2] * time + coeffs[3];
}

double temperatureInterpolate(float *temperature, float time, float longitude)
{
    int prevslot, nextslot;

    float timeFrac;

    double locTime = time - longitude * rad2deg / 15.;

    double timeInterval = 24. / numTemperatures();

    if (locTime < 0.)
        locTime += 24;
    if (locTime > 24)
        locTime -= 24;

    prevslot = (int)(lrint(floor(locTime)) / timeInterval);
    nextslot = prevslot + 1;
    if (nextslot == numTemperatures())
        nextslot = 0;

    timeFrac = locTime - timeInterval * prevslot;

    return temperature[prevslot] +
        (timeFrac / timeInterval) * (temperature[nextslot] -
                                     temperature[prevslot]);


}


int main(int argc, char *argv[])
{
    int angleloss = 0;

    double singleSlope;

    double singleAspect;

    double singleAlbedo;

    double singleLinke;


    struct GModule *module;

    struct
    {
        struct Option *elevin, *aspin, *aspect, *slopein, *slope, *linkein,
            *lin, *albedo, *longin, *alb, *latin, *coefbh, *coefdh,
            *beam_rad, *diff_rad, *refl_rad, *glob_pow, *day, *step, *declin,
            *ltime, *dist, *trackAxes, *coeffTemp,
            *horizon, *horizonstep, *numPartitions, *civilTime,
            *modelparameters;
    }
    parm;

    struct
    {
        struct Flag *shade, *saveMemory, *temperaturecorr, *angleLoss;
    }
    flag;

    struct GridGeometry gridGeom;

    setAngularLossDenominator();

    G_gisinit(argv[0]);
    module = G_define_module();

    module->description =
        "Computes direct (beam), diffuse and reflected solar irradiation raster "
        "maps for given day, latitude, surface and atmospheric conditions. Solar "
        "parameters (e.g. sunrise, sunset times, declination, extraterrestrial "
        "irradiance, daylight length) are saved in a local text file. "
        "Alternatively, a local time can be specified to compute solar "
        "incidence angle and/or irradiance raster maps. The shadowing effect of "
        "the topography is optionally incorporated. ";

    G_get_set_window(&cellhd);
    gridGeom.stepx = cellhd.ew_res;
    gridGeom.stepy = cellhd.ns_res;
    invstepx = 1. / gridGeom.stepx;
    invstepy = 1. / gridGeom.stepy;
    n /*n_cols */  = cellhd.cols;
    m /*n_rows */  = cellhd.rows;
    xmin = cellhd.west;
    ymin = cellhd.south;
    xmax = cellhd.east;
    ymax = cellhd.north;
    gridGeom.deltx = fabs(cellhd.east - cellhd.west);
    gridGeom.delty = fabs(cellhd.north - cellhd.south);

    parm.elevin = G_define_option();
    parm.elevin->key = "elevation";
    parm.elevin->type = TYPE_STRING;
    parm.elevin->required = YES;
    parm.elevin->gisprompt = "old,cell,raster";
    parm.elevin->description = "Name of the elevation raster file";

    parm.aspin = G_define_option();
    parm.aspin->key = "aspect";
    parm.aspin->type = TYPE_STRING;
    parm.aspin->required = NO;
    parm.aspin->gisprompt = "old,cell,raster";
    parm.aspin->description = "Name of the aspect raster file";

    parm.aspect = G_define_option();
    parm.aspect->key = "aspect_value";
    parm.aspect->type = TYPE_DOUBLE;
    parm.aspect->answer = ASPECT;
    parm.aspect->required = NO;
    parm.aspect->description =
        "A single value of the orientation (aspect), 270 is south";

    parm.slopein = G_define_option();
    parm.slopein->key = "slope";
    parm.slopein->type = TYPE_STRING;
    parm.slopein->required = NO;
    /*      parm.slopein->gisprompt = "old,cell,raster";
     */
    parm.slopein->description = "Name of the slope raster file";

    parm.slope = G_define_option();
    parm.slope->key = "slope_value";
    parm.slope->type = TYPE_DOUBLE;
    parm.slope->answer = SLOPE;
    parm.slope->required = NO;
    parm.slope->description = "A single value of inclination (slope)";

    parm.linkein = G_define_option();
    parm.linkein->key = "linke";
    parm.linkein->type = TYPE_STRING;
    parm.linkein->required = NO;
    parm.linkein->gisprompt = "old,cell,raster";
    parm.linkein->description =
        "Name of the Linke turbidity coefficient raster file";

    parm.lin = G_define_option();
    parm.lin->key = "linke_value";
    parm.lin->type = TYPE_DOUBLE;
    parm.lin->answer = LINKE;
    parm.lin->required = NO;
    parm.lin->description =
        "A single value of the Linke turbidity coefficient";

    parm.trackAxes = G_define_option();
    parm.trackAxes->key = "numaxes";
    parm.trackAxes->type = TYPE_INTEGER;
    parm.trackAxes->required = NO;
    parm.trackAxes->description = "Number of tracking type (0 - 5)";


    parm.albedo = G_define_option();
    parm.albedo->key = "albedo";
    parm.albedo->type = TYPE_STRING;
    parm.albedo->required = NO;
    parm.albedo->gisprompt = "old,cell,raster";
    parm.albedo->description = "Name of the albedo coefficient raster file";

    parm.alb = G_define_option();
    parm.alb->key = "albedo_value";
    parm.alb->type = TYPE_DOUBLE;
    parm.alb->answer = ALB;
    parm.alb->required = NO;
    parm.alb->description = "A single value of the albedo coefficient";

    parm.latin = G_define_option();
    parm.latin->key = "lat";
    parm.latin->type = TYPE_STRING;
    parm.latin->required = NO;
    parm.latin->gisprompt = "old,cell,raster";
    parm.latin->description = "Name of the latitude raster file";


    parm.longin = G_define_option();
    parm.longin->key = "longin";
    parm.longin->type = TYPE_STRING;
    parm.longin->required = NO;
    parm.longin->gisprompt = "old,cell,raster";
    parm.longin->description = "Name of the longitude raster file";


    parm.coefbh = G_define_option();
    parm.coefbh->key = "coefbh";
    parm.coefbh->type = TYPE_STRING;
    parm.coefbh->required = NO;
    parm.coefbh->gisprompt = "old,cell,raster";
    parm.coefbh->description = "The real-sky beam radiation coefficient file";

    parm.coefdh = G_define_option();
    parm.coefdh->key = "coefdh";
    parm.coefdh->type = TYPE_STRING;
    parm.coefdh->required = NO;
    parm.coefdh->gisprompt = "old,cell,raster";
    parm.coefdh->description =
        "The real-sky diffuse radiation coefficient file";


    parm.horizon = G_define_option();
    parm.horizon->key = "horizon_basename";
    parm.horizon->type = TYPE_STRING;
    parm.horizon->required = NO;
    parm.horizon->gisprompt = "old,cell,raster";
    parm.horizon->description = "The horizon information file prefix";

    parm.horizonstep = G_define_option();
    parm.horizonstep->key = "horizon_step";
    parm.horizonstep->type = TYPE_DOUBLE;
    parm.horizonstep->required = NO;
    parm.horizonstep->description =
        "Angle step size for the horizon information (degrees)";

    parm.beam_rad = G_define_option();
    parm.beam_rad->key = "beam_rad";
    parm.beam_rad->type = TYPE_STRING;
    parm.beam_rad->required = NO;
    parm.beam_rad->gisprompt = "old,cell,raster";
    parm.beam_rad->description =
        "Output direct (beam) irradiance/irradiation file (raster)";

    parm.diff_rad = G_define_option();
    parm.diff_rad->key = "diff_rad";
    parm.diff_rad->type = TYPE_STRING;
    parm.diff_rad->required = NO;
    parm.diff_rad->gisprompt = "old,cell,raster";
    parm.diff_rad->description =
        "Output diffuse irradiance/irradiation file (raster)";

    parm.refl_rad = G_define_option();
    parm.refl_rad->key = "refl_rad";
    parm.refl_rad->type = TYPE_STRING;
    parm.refl_rad->required = NO;
    parm.refl_rad->gisprompt = "old,cell,raster";
    parm.refl_rad->description =
        "Output reflected irradiance/irradiation file (raster)";

    parm.glob_pow = G_define_option();
    parm.glob_pow->key = "glob_pow";
    parm.glob_pow->type = TYPE_STRING;
    parm.glob_pow->required = NO;
    parm.glob_pow->gisprompt = "old,cell,raster";
    parm.glob_pow->description =
        "Output file (raster) with global irradiance/irradiation or PV power/energy";


    parm.coeffTemp = G_define_option();
    parm.coeffTemp->key = "temperatures";
    parm.coeffTemp->type = TYPE_STRING;
    parm.coeffTemp->required = NO;
    parm.coeffTemp->multiple = YES;
    parm.coeffTemp->gisprompt = "old,cell,raster";
    parm.coeffTemp->description = "Name(s) of the temperature raster map(s)";


    parm.day = G_define_option();
    parm.day->key = "day";
    parm.day->type = TYPE_INTEGER;
    parm.day->required = YES;
    parm.day->description = "No. of day of the year (1-365)";

    parm.step = G_define_option();
    parm.step->key = "step";
    parm.step->type = TYPE_DOUBLE;
    parm.step->answer = STEP;
    parm.step->required = NO;
    parm.step->description = "Time step computing all-day radiation";

    parm.declin = G_define_option();
    parm.declin->key = "declin";
    parm.declin->type = TYPE_DOUBLE;
    parm.declin->required = NO;
    parm.declin->description =
        "Required declination value (overriding the internal value)";

    parm.ltime = G_define_option();
    parm.ltime->key = "time";
    parm.ltime->type = TYPE_DOUBLE;
    /*          parm.ltime->answer = TIME; */
    parm.ltime->required = NO;
    parm.ltime->description = "Local (solar) time [decimal hours]";


    parm.dist = G_define_option();
    parm.dist->key = "dist";
    parm.dist->type = TYPE_DOUBLE;
    parm.dist->answer = DIST;
    parm.dist->required = NO;
    parm.dist->description = "Sampling distance step coefficient (0.5-1.5)";

    parm.numPartitions = G_define_option();
    parm.numPartitions->key = "numpartitions";
    parm.numPartitions->type = TYPE_INTEGER;
    parm.numPartitions->answer = NUM_PARTITIONS;
    parm.numPartitions->required = NO;
    parm.numPartitions->description =
        "Read the input files in this number of chunks";

    parm.civilTime = G_define_option();
    parm.civilTime->key = "civiltime";
    parm.civilTime->type = TYPE_DOUBLE;
    parm.civilTime->required = NO;
    parm.civilTime->description =
        "(optional) The civil time zone value, if none, the time will be local solar time";

    parm.modelparameters = G_define_option();
    parm.modelparameters->key = "modelparameters";
    parm.modelparameters->type = TYPE_STRING;
    parm.modelparameters->required = NO;
    parm.modelparameters->description =
        "(optional) Name of the file with the parameters for the power rating model (file should be in local directory)";



    flag.shade = G_define_flag();
    flag.shade->key = 's';
    flag.shade->description =
        "Do you want to incorporate the shadowing effect of terrain (y/n)";

    flag.saveMemory = G_define_flag();
    flag.saveMemory->key = 'm';
    flag.saveMemory->description =
        "Do you want to use the low-memory version of the program (y/n)";

    flag.temperaturecorr = G_define_flag();
    flag.temperaturecorr->key = 't';
    flag.temperaturecorr->description =
        "Incorporate c-Si temperature effects? (y/n)";

    flag.angleLoss = G_define_flag();
    flag.angleLoss->key = 'a';
    flag.angleLoss->description =
        "Do you want to include the effect of shallow angle reflectivity (y/n)";

    if (G_parser(argc, argv))
        exit(1);

    modelparameters = parm.modelparameters->answer;

    initEfficiencyCoeffs(modelparameters);


    setUseShadow(flag.shade->answer);
    saveMemory = flag.saveMemory->answer;
    angleloss = flag.angleLoss->answer;
    civiltime = parm.civilTime->answer;

    highIrr = 1;

    useTemperatureCorr = flag.temperaturecorr->answer;
    if (useTemperatureCorr) {
        coeftemp = parm.coeffTemp->answers;
    }

    elevin = parm.elevin->answer;
    aspin = parm.aspin->answer;
    slopein = parm.slopein->answer;
    linkein = parm.linkein->answer;
    albedo = parm.albedo->answer;
    latin = parm.latin->answer;
    numaxestracking = parm.trackAxes->answer;


    if (civiltime != NULL) {
        setUseCivilTime(1);
        longin = parm.longin->answer;
        sscanf(parm.civilTime->answer, "%lf", &civilTime);

        // Normalize if somebody should be weird enough to give more than +- 12 
        // hours offset.

        //        civilTime -= 24 * ((int)civilTime / 24.);
        if (civilTime < -12.) {
            civilTime += 24.;
        }
        else if (civilTime > 12.) {
            civilTime -= 24;
        }

    }

    coefbh = parm.coefbh->answer;
    coefdh = parm.coefdh->answer;
    beam_rad = parm.beam_rad->answer;
    horizon = parm.horizon->answer;
    setUseHorizonData(horizon != NULL);
    diff_rad = parm.diff_rad->answer;
    refl_rad = parm.refl_rad->answer;
    glob_pow = parm.glob_pow->answer;




    sscanf(parm.day->answer, "%d", &day);


    if (parm.step->answer != NULL) {
        if (sscanf(parm.step->answer, "%lf", &step) != 1) {
            G_fatal_error("Error reading time step size");

        }
    }
    if (parm.horizonstep->answer != NULL) {
        str_step = parm.horizonstep->answer;

        if (sscanf(parm.horizonstep->answer, "%lf", &horizonStep) != 1) {
            G_fatal_error("Error reading horizon step size");

        }
        setHorizonInterval(deg2rad * horizonStep);
    }


    tt = parm.ltime->answer;
    if (parm.ltime->answer != NULL) {
        fprintf(stdout,
                "Mode 1: instantaneous solar incidence angle & irradiance using a set local time\n");
        fflush(stdout);
        sscanf(parm.ltime->answer, "%lf", &timo);
    }
    else {
        fprintf(stdout, "Mode 2: integrated daily irradiation\n");
        fflush(stdout);
    }


    if (parm.linkein->answer == NULL)
        sscanf(parm.lin->answer, "%lf", &singleLinke);
    if (parm.albedo->answer == NULL)
        sscanf(parm.alb->answer, "%lf", &singleAlbedo);
    /*
       if (parm.lat->answer != NULL)
       sscanf(parm.lat->answer, "%lf", &latitude);
     */
    if (parm.slopein->answer == NULL)
        sscanf(parm.slope->answer, "%lf", &singleSlope);
    singleSlope *= deg2rad;

    if (parm.trackAxes->answer != NULL)
        sscanf(parm.trackAxes->answer, "%d", &numAxisTracking);

    if ((numAxisTracking < 0) || (numAxisTracking > 5)) {
        printf("The number of of the tracking strategy \n");
        printf("cannot be less than 0 or greater than 5.\n");
        exit(1);


    }

    if (parm.aspin->answer == NULL)
        sscanf(parm.aspect->answer, "%lf", &singleAspect);
    singleAspect *= deg2rad;

    if (parm.coefbh->answer == NULL)
        cbh = BSKY;
    if (parm.coefdh->answer == NULL)
        cdh = DSKY;
    sscanf(parm.dist->answer, "%lf", &dist);

    if (parm.numPartitions->answer != NULL) {
        sscanf(parm.numPartitions->answer, "%d", &numPartitions);
        if (useShadow() && (!useHorizonData()) && (numPartitions != 1)) {
            /* If you calculate shadows on the fly, the number of partitions
               must be one.
             */
            G_fatal_error
                ("If you use -s and no horizon rasters, numpartitions must be =1");

        }
    }


    gridGeom.stepxy = dist * 0.5 * (gridGeom.stepx + gridGeom.stepy);
    TOLER = gridGeom.stepxy * EPS;

    /* The save memory scheme will not work if you want to calculate shadows
       on the fly. If you calculate without shadow effects or if you have the
       shadows pre-calculated, there is no problem. */


    if (parm.declin->answer == NULL)
        declination = com_declin(day);
    else {
        sscanf(parm.declin->answer, "%lf", &declin);
        declination = -declin;
    }

    /*
       if (lt != NULL)
       latitude = -latitude * deg2rad;
     */
    if (tt != 0) {
        /* Shadow for just one time during the day */
        if (horizon == NULL) {
            arrayNumInt = 1;
        }
        else {
            arrayNumInt = (int)(360. / horizonStep);

        }
    }
    else {
        /*        Number of bytes holding the horizon information */
        arrayNumInt = (int)(360. / horizonStep);

    }

    if (tt != NULL) {

        tim = (timo - 12) * 15;
        /* converting to degrees */
        /* Jenco (12-timeAngle) * 15 */
        if (tim < 0)
            tim += 360;
        tim = deg2rad * tim;
        /* conv. to radians */
    }

    /* Set up parameters for projection to lat/long if necessary */


    struct Key_Value *in_proj_info, *in_unit_info;

    struct Key_Value *out_proj_info, *out_unit_info;

    double a, es;

    char buf[50];

    if ((in_proj_info = G_get_projinfo()) == NULL)
        G_fatal_error
            ("Can't get projection info of current location: please set latitude via 'lat' or 'latin' option!");

    if ((in_unit_info = G_get_projunits()) == NULL)
        G_fatal_error("Can't get projection units of current location");

    if (pj_get_kv(&iproj, in_proj_info, in_unit_info) < 0)
        G_fatal_error("Can't get projection key values of current location");

    /* Set output projection to latlong w/ same ellipsoid */
    out_proj_info = G_create_key_value();
    out_unit_info = G_create_key_value();

    G_set_key_value("proj", "ll", out_proj_info);

    G_get_ellipsoid_parameters(&a, &es);
    sprintf(buf, "%.16g", a);
    G_set_key_value("a", buf, out_proj_info);
    sprintf(buf, "%.16g", es);
    G_set_key_value("es", buf, out_proj_info);

    G_set_key_value("unit", "degree", out_unit_info);
    G_set_key_value("units", "degrees", out_unit_info);
    G_set_key_value("meters", "1.0", out_unit_info);

    if (pj_get_kv(&oproj, out_proj_info, out_unit_info) < 0)
        G_fatal_error("Unable to set up lat/long projection parameters");

    G_free_key_value(in_proj_info);
    G_free_key_value(in_unit_info);
    G_free_key_value(out_proj_info);
    G_free_key_value(out_unit_info);

/**********end of parser - ******************************/

    calculate(angleloss, singleSlope, singleAspect, singleAlbedo, gridGeom);
    OUTGR();

    return 1;
}


int INPUT_part(int offset)
{
    int h;

    size_t decimals;

    int finalRow, rowrevoffset;

    int loc_num_temperatures;

    int numRows;

    double angle_deg = 0.;

    FCELL *cell1 = NULL, *cell2 = NULL;

    FCELL *cell3 = NULL, *cell4 = NULL, *cell5 = NULL, *cell6 = NULL, *cell7 =
        NULL;
    FCELL *rast1 = NULL, *rast2 = NULL;

    FCELL *rasttemp = NULL;

    static FCELL **horizonbuf;

    unsigned char *horizonpointer;

    int fd1 = -1, fd2 = -1, fd3 = -1, fd4 = -1, fd5 = -1, fd6 = -1, fd7 =
        -1, row, row_rev;
    int ft0 = -1, ft1 = -1, ft2 = -1, ft3 = -1;

    static int *fd_shad;

    int fr1 = -1, fr2 = -1;

    int l, i, j;

    char *shad_filename;

    char tempcoefname[256];

    finalRow = m - offset - m / numPartitions;
    if (finalRow < 0) {
        finalRow = 0;
    }

    numRows = m / numPartitions;

    cell1 = Rast_allocate_f_buf();

    if (z == NULL) {
        z = (float **)malloc(sizeof(float *) * (numRows));


        for (l = 0; l < numRows; l++) {
            z[l] = (float *)malloc(sizeof(float) * (n));
        }
    }




    /*
       if((mapset=G_find_cell(elevin,""))==NULL)
       printf("cell file not found\n");
     */


    fd1 = Rast_open_old(elevin, "");

    if (slopein != NULL) {
        cell3 = Rast_allocate_f_buf();
        if (s == NULL) {
            s = (float **)malloc(sizeof(float *) * (numRows));

            for (l = 0; l < numRows; l++) {
                s[l] = (float *)malloc(sizeof(float) * (n));
            }

        }
        fd3 = Rast_open_old(slopein, "");

    }

    if (aspin != NULL) {
        cell2 = Rast_allocate_f_buf();

        if (o == NULL) {
            o = (float **)malloc(sizeof(float *) * (numRows));

            for (l = 0; l < numRows; l++) {
                o[l] = (float *)malloc(sizeof(float) * (n));
            }
        }

        fd2 = Rast_open_old(aspin, "");

    }


    if (linkein != NULL) {
        cell4 = Rast_allocate_f_buf();
        if (li == NULL) {
            li = (float **)malloc(sizeof(float *) * (numRows));
            for (l = 0; l < numRows; l++)
                li[l] = (float *)malloc(sizeof(float) * (n));

        }

        fd4 = Rast_open_old(linkein, "");
    }

    if (albedo != NULL) {
        cell5 = Rast_allocate_f_buf();
        if (a == NULL) {
            a = (float **)malloc(sizeof(float *) * (numRows));
            for (l = 0; l < numRows; l++)
                a[l] = (float *)malloc(sizeof(float) * (n));
        }

        fd5 = Rast_open_old(albedo, "");
    }

    if (latin != NULL) {
        cell6 = Rast_allocate_f_buf();
        if (la == NULL) {
            la = (float **)malloc(sizeof(float *) * (numRows));
            for (l = 0; l < numRows; l++)
                la[l] = (float *)malloc(sizeof(float) * (n));
        }

        fd6 = Rast_open_old(latin, "");
    }

    if (longin != NULL) {
        cell7 = Rast_allocate_f_buf();
        longitArray = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++)
            longitArray[l] = (float *)malloc(sizeof(float) * (n));


        fd7 = Rast_open_old(longin, "");
    }



    if (coefbh != NULL) {
        rast1 = Rast_allocate_f_buf();

        if (cbhr == NULL) {
            cbhr = (float **)malloc(sizeof(float *) * (numRows));
            for (l = 0; l < numRows; l++)
                cbhr[l] = (float *)malloc(sizeof(float) * (n));
        }

        fr1 = Rast_open_old(coefbh, "");
    }

    if (coefdh != NULL) {
        rast2 = Rast_allocate_f_buf();
        if (cdhr == NULL) {
            cdhr = (float **)malloc(sizeof(float *) * (numRows));
            for (l = 0; l < numRows; l++)
                cdhr[l] = (float *)malloc(sizeof(float) * (n));
        }

        fr2 = Rast_open_old(coefdh, "");
    }

    if (coeftemp != NULL) {

        for (i = 0; coeftemp[i]; i++) ;
        setNumTemperatures(i);
        loc_num_temperatures = i;


        rasttemp = Rast_allocate_f_buf();
        if (tempdata == NULL) {
            tempdata =
                (float *)malloc(sizeof(float *) *
                                (numRows * n * num_temperatures));

        }


        for (i = 0; i < loc_num_temperatures; i++) {


            ft1 = Rast_open_old(coeftemp[i], "");
            for (row = m - offset - 1; row >= finalRow; row--) {
                row_rev = m - row - 1;
                rowrevoffset = row_rev - offset;
                Rast_get_f_row(ft1, rasttemp, row);
                for (j = 0; j < n; j++) {
                    if (!Rast_is_f_null_value(rasttemp + j))
                        tempdata[(rowrevoffset * n +
                                  j) * loc_num_temperatures + i] =
                            (float)rasttemp[j];
                    else
                        tempdata[(rowrevoffset * n +
                                  j) * loc_num_temperatures + i] = UNDEFZ;
                }
            }
            Rast_close(ft1);
        }
        G_free(rasttemp);
    }


    if (useHorizonData()) {
        if (horizonarray == NULL) {
            horizonarray =
                (unsigned char *)malloc(sizeof(char) * arrayNumInt * numRows *
                                        n);

            horizonbuf = (FCELL **) malloc(sizeof(FCELL *) * arrayNumInt);
            fd_shad = (int *)malloc(sizeof(int) * arrayNumInt);
        }
        decimals = G_get_num_decimals(str_step);
        angle_deg = 0;
        for (i = 0; i < arrayNumInt; i++) {
            horizonbuf[i] = Rast_allocate_f_buf();
            shad_filename = G_generate_basename(horizon, angle_deg,
                                                3, decimals);
            fd_shad[i] = Rast_open_old(shad_filename, "");
            angle_deg += horizonStep;
            G_free(shad_filename);
        }
    }

    if (horizon != NULL) {


        for (i = 0; i < arrayNumInt; i++) {
            for (row = m - offset - 1; row >= finalRow; row--) {

                row_rev = m - row - 1;
                rowrevoffset = row_rev - offset;
                Rast_get_f_row(fd_shad[i], horizonbuf[i], row);
                horizonpointer =
                    horizonarray + arrayNumInt * n * rowrevoffset;
                for (j = 0; j < n; j++) {

                    horizonpointer[i] = (char)(rint(SCALING_FACTOR
                                                    * fmin(horizonbuf[i][j],
                                                           256 * invScale)));
                    horizonpointer += arrayNumInt;

                }
            }
        }


    }



    for (row = m - offset - 1; row >= finalRow; row--) {
        Rast_get_f_row(fd1, cell1, row);
        if (aspin != NULL)
            Rast_get_f_row(fd2, cell2, row);
        if (slopein != NULL)
            Rast_get_f_row(fd3, cell3, row);
        if (linkein != NULL)
            Rast_get_f_row(fd4, cell4, row);
        if (albedo != NULL)
            Rast_get_f_row(fd5, cell5, row);
        if (latin != NULL)
            Rast_get_f_row(fd6, cell6, row);
        if (longin != NULL)
            Rast_get_f_row(fd7, cell7, row);
        if (coefbh != NULL)
            Rast_get_f_row(fr1, rast1, row);
        if (coefdh != NULL)
            Rast_get_f_row(fr2, rast2, row);


        row_rev = m - row - 1;
        rowrevoffset = row_rev - offset;

        for (j = 0; j < n; j++) {
            if (!Rast_is_f_null_value(cell1 + j))
                z[rowrevoffset][j] = (float)cell1[j];
            else
                z[rowrevoffset][j] = UNDEFZ;

            if (aspin != NULL) {
                if (!Rast_is_f_null_value(cell2 + j))
                    o[rowrevoffset][j] = (float)cell2[j];
                else
                    o[rowrevoffset][j] = UNDEFZ;
            }
            if (slopein != NULL) {
                if (!Rast_is_f_null_value(cell3 + j))
                    s[rowrevoffset][j] = (float)cell3[j];
                else
                    s[rowrevoffset][j] = UNDEFZ;
            }

            if (linkein != NULL) {
                if (!Rast_is_f_null_value(cell4 + j))
                    li[rowrevoffset][j] = (float)cell4[j];
                else
                    li[rowrevoffset][j] = UNDEFZ;
            }

            if (albedo != NULL) {
                if (!Rast_is_f_null_value(cell5 + j))
                    a[rowrevoffset][j] = (float)cell5[j];
                else
                    a[rowrevoffset][j] = UNDEFZ;
            }

            if (latin != NULL) {
                if (!Rast_is_f_null_value(cell6 + j))
                    la[rowrevoffset][j] = (float)cell6[j];
                else
                    la[rowrevoffset][j] = UNDEFZ;
            }

            if (longin != NULL) {
                if (!Rast_is_f_null_value(cell7 + j))
                    longitArray[rowrevoffset][j] = (float)cell7[j];
                else
                    longitArray[rowrevoffset][j] = UNDEFZ;
            }

            if (coefbh != NULL) {
                if (!Rast_is_f_null_value(rast1 + j))
                    cbhr[rowrevoffset][j] = (float)rast1[j];
                else
                    cbhr[rowrevoffset][j] = UNDEFZ;
            }

            if (coefdh != NULL) {
                if (!Rast_is_f_null_value(rast2 + j))
                    cdhr[rowrevoffset][j] = (float)rast2[j];
                else
                    cdhr[rowrevoffset][j] = UNDEFZ;
            }



        }
    }

    Rast_close(fd1);
    G_free(cell1);

    if (aspin != NULL) {
        G_free(cell2);
        Rast_close(fd2);
    }
    if (slopein != NULL) {
        G_free(cell3);
        Rast_close(fd3);
    }
    if (linkein != NULL) {
        G_free(cell4);
        Rast_close(fd4);
    }
    if (albedo != NULL) {
        G_free(cell5);
        Rast_close(fd5);
    }
    if (latin != NULL) {
        G_free(cell6);
        Rast_close(fd6);
    }
    if (longin != NULL) {
        G_free(cell7);
        Rast_close(fd7);
    }
    if (coefbh != NULL) {
        G_free(rast1);
        Rast_close(fr1);
    }
    if (coefdh != NULL) {
        G_free(rast2);
        Rast_close(fr2);
    }



    if (useHorizonData()) {
        for (i = 0; i < arrayNumInt; i++) {
            Rast_close(fd_shad[i]);
            G_free(horizonbuf[i]);
        }
    }


/*******transformation of angles from 0 to east counterclock
        to 0 to north clocwise, for ori=0 upslope flowlines
        turn the orientation 2*M_PI ************/

    /* needs to be eliminated */

    /*for (i = 0; i < m; ++i) */
    for (i = 0; i < numRows; i++) {
        for (j = 0; j < n; j++) {
            zmax = AMAX1(zmax, z[i][j]);
            if (aspin != NULL) {
                if (o[i][j] != 0.) {
                    if (o[i][j] < 90.)
                        o[i][j] = 90. - o[i][j];
                    else
                        o[i][j] = 450. - o[i][j];
                }
                /*   printf("o,z = %d  %d i,j, %d %d \n", o[i][j],z[i][j],i,j); */

                if ((aspin != NULL) && o[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if ((slopein != NULL) && s[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if (linkein != NULL && li[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if (albedo != NULL && a[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if (latin != NULL && la[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if (coefbh != NULL && cbhr[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if (coefdh != NULL && cdhr[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
            }

        }
    }

    return 1;
}

int OUTGR(void)
{
    FCELL *cell8 = NULL, *cell9 = NULL, *cell10 = NULL, *cell12 = NULL;

    int fd8 = -1, fd9 = -1, fd10 = -1, fd12 = -1;

    int i, iarc, j;

    char msg[100];



    if (beam_rad != NULL) {
        cell8 = Rast_allocate_f_buf();
        fd8 = Rast_open_fp_new(beam_rad);
        if (fd8 < 0) {
            G_fatal_error("unable to create raster map %s", beam_rad);
            exit(1);
        }
    }


    if (diff_rad != NULL) {
        cell9 = Rast_allocate_f_buf();
        fd9 = Rast_open_fp_new(diff_rad);
        if (fd9 < 0) {
            G_fatal_error("unable to create raster map %s", diff_rad);
            exit(1);
        }
    }

    if (refl_rad != NULL) {
        cell10 = Rast_allocate_f_buf();
        fd10 = Rast_open_fp_new(refl_rad);
        if (fd10 < 0) {
            G_fatal_error("unable to create raster map %s", refl_rad);
            exit(1);
        }
    }

    if (glob_pow != NULL) {
        cell12 = Rast_allocate_f_buf();
        fd12 = Rast_open_fp_new(glob_pow);
        if (fd12 < 0) {
            G_fatal_error("unable to create raster map %s", glob_pow);
            exit(1);
        }
    }



    if (m != Rast_window_rows()) {
        fprintf(stderr, "OOPS: rows changed from %d to %d\n", m,
                Rast_window_rows());
        exit(1);
    }
    if (n != Rast_window_cols()) {
        fprintf(stderr, "OOPS: cols changed from %d to %d\n", n,
                Rast_window_cols());
        exit(1);
    }

    for (iarc = 0; iarc < m; iarc++) {
        i = m - iarc - 1;

        if (beam_rad != NULL) {
            for (j = 0; j < n; j++) {
                if (beam[i][j] == UNDEFZ)
                    Rast_set_f_null_value(cell8 + j, 1);
                else
                    cell8[j] = (FCELL) beam[i][j];

            }
            Rast_put_f_row(fd8, cell8);
        }

        if (glob_pow != NULL) {
            for (j = 0; j < n; j++) {
                if (globrad[i][j] == UNDEFZ)
                    Rast_set_f_null_value(cell12 + j, 1);
                else
                    cell12[j] = (FCELL) globrad[i][j];

            }
            Rast_put_f_row(fd12, cell12);
        }



        if (diff_rad != NULL) {
            for (j = 0; j < n; j++) {
                if (diff[i][j] == UNDEFZ)
                    Rast_set_f_null_value(cell9 + j, 1);
                else
                    cell9[j] = (FCELL) diff[i][j];
            }
            Rast_put_f_row(fd9, cell9);
        }

        if (refl_rad != NULL) {
            for (j = 0; j < n; j++) {
                if (refl[i][j] == UNDEFZ)
                    Rast_set_f_null_value(cell10 + j, 1);
                else
                    cell10[j] = (FCELL) refl[i][j];
            }
            Rast_put_f_row(fd10, cell10);
        }

    }

    if (beam_rad != NULL) {
        Rast_close(fd8);
        Rast_write_history(beam_rad, &hist);
    }
    if (diff_rad != NULL) {
        Rast_close(fd9);
        Rast_write_history(diff_rad, &hist);
    }
    if (refl_rad != NULL) {
        Rast_close(fd10);
        Rast_write_history(refl_rad, &hist);
    }
    if (glob_pow != NULL) {
        Rast_close(fd12);
        Rast_write_history(glob_pow, &hist);
    }

    return 1;
}

/*  min(), max() are unused
   int min(arg1, arg2)
   int arg1;
   int arg2;
   {
   int res;
   if (arg1 <= arg2) {
   res = arg1;
   }
   else {
   res = arg2;
   }
   return res;
   }

 */

/**********************************************************/



#define C1 1.15
#define C2 -0.32
#define P_STC 94.804
#define V_STC 34.6
#define I_STC 2.74
#define ALPHA_VM 0.0033
#define BETA_VM -0.159
#define T_STC 25.
#define NOCT 48.

double modelConstants[8];

int initEfficiencyCoeffs(char *filename)
{

    int i;

    FILE *fp;


    modelConstants[0] = 94.804;
    modelConstants[1] = 3.151;
    modelConstants[2] = -0.8768;
    modelConstants[3] = -0.32148;
    modelConstants[4] = 0.003795;
    modelConstants[5] = -0.001056;
    modelConstants[6] = -0.0005247;

    /* This is the heating coefficient for the modules */
    modelConstants[7] = 0.035;

    if (!filename)
        return 0;

    fp = fopen(filename, "r");

    if (!fp) {
        printf("Could not open coefficients file %s. Aborting.\n", filename);
        exit(1);
    }
    for (i = 0; i < 8; i++) {
        if (fscanf(fp, "%lf", modelConstants + i) != 1) {
            printf
                ("Could not read coefficient from coefficients file. Aborting.\n");
            exit(1);
        }
    }
    fclose(fp);

    return 0;

}


double efficiency(double irr, double temp)
{
    double pm;

    double relirr, lnrelirr;

    double tprime;

    relirr = 0.001 * irr;

    if (relirr <= 0.)
        return 0.;

    lnrelirr = log(relirr);

    temp = irr * modelConstants[7] + temp;

    tprime = temp - T_STC;

    pm = modelConstants[0] + lnrelirr * (modelConstants[1] +
                                         lnrelirr * modelConstants[2]) +
        tprime * (modelConstants[3] +
                  lnrelirr * (modelConstants[4] +
                              lnrelirr * modelConstants[5]) +
                  modelConstants[6] * tprime);

    /*
       im = I_STC + ALPHA_VM*tprime;

       vm = V_STC + C1*lnrelirr + C2*lnrelirr*lnrelirr + BETA_VM*tprime;
       return im*vm/P_STC;  
     */
    return pm / modelConstants[0];


}


double joules2(struct SunGeometryConstDay *sunGeom,
               struct SunGeometryVarDay *sunVarGeom,
               struct SunGeometryVarSlope *sunSlopeGeom,
               struct SolarRadVar *sunRadVar,
               struct GridGeometry *gridGeom, float *temperatureData,
               unsigned char *horizonpointer,
               double latitude, double longitude, double sinAxisSlope,
               double cosAxisSlope, BeamRadFunc bRadFunc,
               DiffRadFunc dRadFunc)
{

    double s0, dfr, dfr_rad;

    double ra, dra;

    int ss = 1;

    double effic = 1.;

    double firstTime;

    double firstAngle, lastAngle;

    double beam_irr, diff_irr, refl_irr;

    double beam_irr_real, diff_irr_real, refl_irr_real;

    double totRad = 0.;

    double totRadCorr = 0.;

    double totrad_real;

    double q1;

    double presTemperature;

    double presTime;

    double latid_l, cos_u, cos_v, sin_u, sin_v;

    double sin_phi_l, tan_lam_l;

    double locAzimuth;

    double sinSolarAzi, cosSolarAzi, sinSolarAlt, cosSolarAlt;

    double axisRotAngle, cosAxisRotAngle;


    double aspect;

    int srStepNo;

    double bh;

    double rr;

    double sinAxisRotAngle, cosAngle, cosAngle2, cosAngle3;

    double newCosAngle, newSinAngle;

    beam_e = 0.;
    diff_e = 0.;
    refl_e = 0.;

    struct SolarRadVar sunRadVar_cs;

    sunRadVar_cs = *sunRadVar;

    if (numAxisTracking == 3) {
        cos_u = cos(pihalf - sunSlopeGeom->slope);      /* = sin(slope) */
        sin_u = sin(pihalf - sunSlopeGeom->slope);      /* = cos(slope) */
    }



    if (highIrr) {
        sunRadVar_cs.cbh = 1.;
        sunRadVar_cs.cdh = 1.;
    }


    com_par(sunGeom, sunVarGeom, gridGeom, latitude, longitude);

    if (tt != NULL) {           /*irradiance */
        presTime = timo - civilTime;


        if (numAxisTracking != 0) {
            if (numAxisTracking == 4) { /* East-west oriented horizontal axis */


                if ((sunVarGeom->solarAzimuth > pihalf)
                    && (sunVarGeom->solarAzimuth < M_PI + pihalf)) {
                    sunSlopeGeom->aspect = M_PI;
                }
                else {
                    sunSlopeGeom->aspect = 0.001;
                }

                sunSlopeGeom->slope =
                    fabs(atan
                         (cos(sunVarGeom->solarAzimuth) /
                          tan(sunVarGeom->solarAltitude)));
                /* Unlike the normal r.sun here the aspect and slope change during the
                   day. */

                cos_v = cos(pihalf + sunSlopeGeom->aspect);
                sin_v = sin(pihalf + sunSlopeGeom->aspect);
                cos_u = cos(pihalf - sunSlopeGeom->slope);      /* = sin(slope) */
                sin_u = sin(pihalf - sunSlopeGeom->slope);      /* = cos(slope) */



            }
            else if (numAxisTracking == 1) {    /* North-south oriented horizontal axis */
                if (sunGeom->timeAngle <= 0.) {
                    aspect = pihalf;
                }
                else {
                    aspect = 1.5 * M_PI;
                }
                sunSlopeGeom->slope =
                    fabs(atan
                         (sin(sunVarGeom->solarAzimuth) /
                          tan(sunVarGeom->solarAltitude)));
                sunSlopeGeom->aspect = aspect;
                /* Unlike the normal r.sun here the aspect and slope change during the
                   day. */

                cos_u = cos(pihalf - sunSlopeGeom->slope);      /* = sin(slope) */
                sin_u = sin(pihalf - sunSlopeGeom->slope);      /* = cos(slope) */
                cos_v = cos(pihalf + sunSlopeGeom->aspect);
                sin_v = sin(pihalf + sunSlopeGeom->aspect);



            }
            else if (numAxisTracking == 3) {    /* Vertical axis */
                sunSlopeGeom->aspect = sunVarGeom->solarAzimuth;
                cos_v = cos(pihalf + sunSlopeGeom->aspect);
                sin_v = sin(pihalf + sunSlopeGeom->aspect);

            }
            else if (numAxisTracking == 5) {    /* Inclined axis */
                locAzimuth = M_PI - sunVarGeom->solarAzimuth;
                sinSolarAzi = sin(locAzimuth);
                cosSolarAzi = cos(locAzimuth);
                sinSolarAlt = sin(sunVarGeom->solarAltitude);
                cosSolarAlt = cos(sunVarGeom->solarAltitude);

                if ((locAzimuth > pihalf)
                    || (locAzimuth < -pihalf)) {
                    axisRotAngle = atan(sinSolarAzi * cosSolarAlt
                                        / (cosSolarAzi * cosSolarAlt *
                                           sinAxisSlope +
                                           sinSolarAlt * cosAxisSlope));
                }
                else {
                    axisRotAngle = atan(sinSolarAzi * cosSolarAlt
                                        / (cosSolarAzi * cosSolarAlt *
                                           sinAxisSlope +
                                           sinSolarAlt * cosAxisSlope));
                }
                if ((axisRotAngle < 0.) && (locAzimuth > 0.)) {
                    axisRotAngle += M_PI;
                }
                else if ((axisRotAngle > 0.) && (locAzimuth < 0.)) {
                    axisRotAngle -= M_PI;
                }
                cosAxisRotAngle = cos(axisRotAngle);

                sunSlopeGeom->slope =
                    pihalf - asin(cosAxisSlope * cosAxisRotAngle);

                if (locAzimuth > 0.) {
                    sunSlopeGeom->aspect
                        =
                        M_PI -
                        acos(sinAxisSlope * cosAxisRotAngle /
                             sin(sunSlopeGeom->slope));
                }
                else {
                    sunSlopeGeom->aspect
                        =
                        M_PI +
                        acos(sinAxisSlope * cosAxisRotAngle /
                             sin(sunSlopeGeom->slope));
                }
                cos_v = cos(pihalf + sunSlopeGeom->aspect);
                sin_v = sin(pihalf + sunSlopeGeom->aspect);
                cos_u = cos(pihalf - sunSlopeGeom->slope);      /* = sin(slope) */
                sin_u = sin(pihalf - sunSlopeGeom->slope);
                /* Debugging: remove the following when not needed anymore */
                sinAxisRotAngle = sin(axisRotAngle);
                cosAngle =
                    cosSolarAzi * cosSolarAlt * sinAxisSlope *
                    cosAxisRotAngle +
                    sinSolarAzi * cosSolarAlt * sinAxisRotAngle +
                    sinSolarAlt * cosAxisSlope * cosAxisRotAngle;

                newCosAngle = cos(axisRotAngle + 0.01);
                newSinAngle = sin(axisRotAngle + 0.01);
                sinAxisRotAngle = sin(axisRotAngle);
                cosAngle2 =
                    cosSolarAzi * cosSolarAlt * sinAxisSlope * newCosAngle +
                    sinSolarAzi * cosSolarAlt * newSinAngle +
                    sinSolarAlt * cosAxisSlope * newCosAngle;
                newCosAngle = cos(axisRotAngle - 0.01);
                newSinAngle = sin(axisRotAngle - 0.01);
                sinAxisRotAngle = sin(axisRotAngle);
                cosAngle3 =
                    cosSolarAzi * cosSolarAlt * sinAxisSlope * newCosAngle +
                    sinSolarAzi * cosSolarAlt * newSinAngle +
                    sinSolarAlt * cosAxisSlope * newCosAngle;

                if ((cosAngle < cosAngle2) || (cosAngle < cosAngle3)) {
                    printf
                        ("Angle not optimal, time=%.2f, azimuth = %f, sunheight = %f, rot. angle = %f\n",
                         presTime, sunVarGeom->solarAzimuth,
                         sunVarGeom->solarAltitude, axisRotAngle);
                }

            }
            else {              /* numAisTracking==2 */

                /*
                   aspect = M_PI +timeAngle;    
                 */
                sunSlopeGeom->aspect = sunVarGeom->solarAzimuth;
                sunSlopeGeom->slope = pihalf - sunVarGeom->solarAltitude;

                /* Unlike the normal r.sun here the aspect and slope change during the
                   day. */

                cos_u = cos(pihalf - sunSlopeGeom->slope);      /* = sin(slope) */
                sin_u = sin(pihalf - sunSlopeGeom->slope);      /* = cos(slope) */
                cos_v = cos(pihalf + sunSlopeGeom->aspect);
                sin_v = sin(pihalf + sunSlopeGeom->aspect);


            }


            sin_phi_l =
                -gridGeom->coslat * cos_u * sin_v + gridGeom->sinlat * sin_u;
            latid_l = asin(sin_phi_l);

            q1 = gridGeom->sinlat * cos_u * sin_v + gridGeom->coslat * sin_u;
            tan_lam_l = -cos_u * cos_v / q1;
            sunSlopeGeom->longit_l = atan(tan_lam_l);

            if ((sunSlopeGeom->aspect < M_PI) &&
                (sunSlopeGeom->longit_l < 0.)) {
                sunSlopeGeom->longit_l += M_PI;
            }
            else if ((sunSlopeGeom->aspect > M_PI) &&
                     (sunSlopeGeom->longit_l > 0.)) {
                sunSlopeGeom->longit_l -= M_PI;
            }
            sunSlopeGeom->lum_C31_l = cos(latid_l) * sunGeom->cosdecl;
            sunSlopeGeom->lum_C33_l = sin_phi_l * sunGeom->sindecl;

        }

        s0 = lumcline2(sunGeom, sunVarGeom, sunSlopeGeom, gridGeom,
                       horizonpointer);

        if (sunVarGeom->solarAltitude > 0.) {
            if ((!sunVarGeom->isShadow) && s0 > 0.) {
                ra = bRadFunc(s0, &bh, sunVarGeom, sunSlopeGeom, sunRadVar);    /* beam radiation */
                beam_e += ra;
            }
            else {
                beam_e = 0.;
                bh = 0.;
            }

            if ((diff_rad != NULL) || (glob_pow != NULL)) {
                dra = drad(s0, bh, &rr, sunVarGeom, sunSlopeGeom, sunRadVar);   /* diffuse rad. */
                diff_e += dra;
            }
            if ((refl_rad != NULL) || (glob_pow != NULL)) {
                if ((diff_rad == NULL) && (glob_pow == NULL))
                    drad(s0, bh, &rr, sunVarGeom, sunSlopeGeom, sunRadVar);
                refl_e += rr;   /* reflected rad. */
            }
            totRad = beam_e + diff_e + refl_e;


            if (useTemperatureCorr) {
                /*
                   presTemperature =
                   temperatureInterpolate(temperatureData, presTime,
                   longitude);
                 */
                effic = efficiency(totRad, *temperatureData);
            }

            totRadCorr = effic * totRad;
        }                       /* solarAltitude */
    }
    else {
        /* all-day radiation */

        srStepNo = (int)(sunGeom->sunrise_time / step);

        if ((sunGeom->sunrise_time - srStepNo * step) > 0.5 * step) {
            firstTime = (srStepNo + 1.5) * step;
        }
        else {
            firstTime = (srStepNo + 0.5) * step;
        }

        presTime = firstTime;

        firstAngle = (firstTime - 12) * HOURANGLE;
        lastAngle = (sunGeom->sunset_time - 12) * HOURANGLE;




        dfr_rad = step * HOURANGLE;

        sunGeom->timeAngle = firstAngle;

        varCount_global = 0;


        dfr = step;

        while (ss == 1) {


            beam_irr = diff_irr = refl_irr = 0.;
            beam_irr_real = diff_irr_real = refl_irr_real = 0.;

            com_par(sunGeom, sunVarGeom, gridGeom, latitude, longitude);


            if (numAxisTracking != 0) {
                if (numAxisTracking == 4) {     /* East-west oriented horizontal axis */


                    if ((sunVarGeom->solarAzimuth > pihalf)
                        && (sunVarGeom->solarAzimuth < M_PI + pihalf)) {
                        sunSlopeGeom->aspect = M_PI;
                    }
                    else {
                        sunSlopeGeom->aspect = 0.001;
                    }

                    sunSlopeGeom->slope =
                        fabs(atan
                             (cos(sunVarGeom->solarAzimuth) /
                              tan(sunVarGeom->solarAltitude)));
                    /* Unlike the normal r.sun here the aspect and slope change during the
                       day. */

                    cos_v = cos(pihalf + sunSlopeGeom->aspect);
                    sin_v = sin(pihalf + sunSlopeGeom->aspect);
                    cos_u = cos(pihalf - sunSlopeGeom->slope);  /* = sin(slope) */
                    sin_u = sin(pihalf - sunSlopeGeom->slope);  /* = cos(slope) */



                }
                else if (numAxisTracking == 1) {        /* North-south oriented horizontal axis */
                    if (sunGeom->timeAngle <= 0.) {
                        aspect = pihalf;
                    }
                    else {
                        aspect = 1.5 * M_PI;
                    }
                    sunSlopeGeom->slope =
                        fabs(atan
                             (sin(sunVarGeom->solarAzimuth) /
                              tan(sunVarGeom->solarAltitude)));
                    sunSlopeGeom->aspect = aspect;
                    /* Unlike the normal r.sun here the aspect and slope change during the
                       day. */

                    cos_u = cos(pihalf - sunSlopeGeom->slope);  /* = sin(slope) */
                    sin_u = sin(pihalf - sunSlopeGeom->slope);  /* = cos(slope) */
                    cos_v = cos(pihalf + sunSlopeGeom->aspect);
                    sin_v = sin(pihalf + sunSlopeGeom->aspect);



                }
                else if (numAxisTracking == 3) {        /* Vertical axis */
                    sunSlopeGeom->aspect = sunVarGeom->solarAzimuth;
                    cos_v = cos(pihalf + sunSlopeGeom->aspect);
                    sin_v = sin(pihalf + sunSlopeGeom->aspect);

                }
                else if (numAxisTracking == 5) {        /* Inclined axis */
                    locAzimuth = M_PI - sunVarGeom->solarAzimuth;
                    sinSolarAzi = sin(locAzimuth);
                    cosSolarAzi = cos(locAzimuth);
                    sinSolarAlt = sin(sunVarGeom->solarAltitude);
                    cosSolarAlt = cos(sunVarGeom->solarAltitude);

                    if ((locAzimuth > pihalf)
                        || (locAzimuth < -pihalf)) {
                        axisRotAngle = atan(sinSolarAzi * cosSolarAlt
                                            / (cosSolarAzi * cosSolarAlt *
                                               sinAxisSlope +
                                               sinSolarAlt * cosAxisSlope));
                    }
                    else {
                        axisRotAngle = atan(sinSolarAzi * cosSolarAlt
                                            / (cosSolarAzi * cosSolarAlt *
                                               sinAxisSlope +
                                               sinSolarAlt * cosAxisSlope));
                    }
                    if ((axisRotAngle < 0.) && (locAzimuth > 0.)) {
                        axisRotAngle += M_PI;
                    }
                    else if ((axisRotAngle > 0.) && (locAzimuth < 0.)) {
                        axisRotAngle -= M_PI;
                    }
                    cosAxisRotAngle = cos(axisRotAngle);

                    sunSlopeGeom->slope =
                        pihalf - asin(cosAxisSlope * cosAxisRotAngle);

                    if (locAzimuth > 0.) {
                        sunSlopeGeom->aspect
                            =
                            M_PI -
                            acos(sinAxisSlope * cosAxisRotAngle /
                                 sin(sunSlopeGeom->slope));
                    }
                    else {
                        sunSlopeGeom->aspect
                            =
                            M_PI +
                            acos(sinAxisSlope * cosAxisRotAngle /
                                 sin(sunSlopeGeom->slope));
                    }
                    cos_v = cos(pihalf + sunSlopeGeom->aspect);
                    sin_v = sin(pihalf + sunSlopeGeom->aspect);
                    cos_u = cos(pihalf - sunSlopeGeom->slope);  /* = sin(slope) */
                    sin_u = sin(pihalf - sunSlopeGeom->slope);
                    /* Debugging: remove the following when not needed anymore */
                    sinAxisRotAngle = sin(axisRotAngle);
                    cosAngle =
                        cosSolarAzi * cosSolarAlt * sinAxisSlope *
                        cosAxisRotAngle +
                        sinSolarAzi * cosSolarAlt * sinAxisRotAngle +
                        sinSolarAlt * cosAxisSlope * cosAxisRotAngle;

                    newCosAngle = cos(axisRotAngle + 0.01);
                    newSinAngle = sin(axisRotAngle + 0.01);
                    sinAxisRotAngle = sin(axisRotAngle);
                    cosAngle2 =
                        cosSolarAzi * cosSolarAlt * sinAxisSlope *
                        newCosAngle +
                        sinSolarAzi * cosSolarAlt * newSinAngle +
                        sinSolarAlt * cosAxisSlope * newCosAngle;
                    newCosAngle = cos(axisRotAngle - 0.01);
                    newSinAngle = sin(axisRotAngle - 0.01);
                    sinAxisRotAngle = sin(axisRotAngle);
                    cosAngle3 =
                        cosSolarAzi * cosSolarAlt * sinAxisSlope *
                        newCosAngle +
                        sinSolarAzi * cosSolarAlt * newSinAngle +
                        sinSolarAlt * cosAxisSlope * newCosAngle;

                    if ((cosAngle < cosAngle2) || (cosAngle < cosAngle3)) {
                        printf
                            ("Angle not optimal, time=%.2f, azimuth = %f, sunheight = %f, rot. angle = %f\n",
                             presTime, sunVarGeom->solarAzimuth,
                             sunVarGeom->solarAltitude, axisRotAngle);
                    }

                }
                else {          /* numAisTracking==2 */

                    /*
                       aspect = M_PI +timeAngle;    
                     */
                    sunSlopeGeom->aspect = sunVarGeom->solarAzimuth;
                    sunSlopeGeom->slope = pihalf - sunVarGeom->solarAltitude;

                    /* Unlike the normal r.sun here the aspect and slope change during the
                       day. */

                    cos_u = cos(pihalf - sunSlopeGeom->slope);  /* = sin(slope) */
                    sin_u = sin(pihalf - sunSlopeGeom->slope);  /* = cos(slope) */
                    cos_v = cos(pihalf + sunSlopeGeom->aspect);
                    sin_v = sin(pihalf + sunSlopeGeom->aspect);


                }


                sin_phi_l =
                    -gridGeom->coslat * cos_u * sin_v +
                    gridGeom->sinlat * sin_u;
                latid_l = asin(sin_phi_l);

                q1 = gridGeom->sinlat * cos_u * sin_v +
                    gridGeom->coslat * sin_u;
                tan_lam_l = -cos_u * cos_v / q1;
                sunSlopeGeom->longit_l = atan(tan_lam_l);

                if ((sunSlopeGeom->aspect < M_PI) &&
                    (sunSlopeGeom->longit_l < 0.)) {
                    sunSlopeGeom->longit_l += M_PI;
                }
                else if ((sunSlopeGeom->aspect > M_PI) &&
                         (sunSlopeGeom->longit_l > 0.)) {
                    sunSlopeGeom->longit_l -= M_PI;
                }
                sunSlopeGeom->lum_C31_l = cos(latid_l) * sunGeom->cosdecl;
                sunSlopeGeom->lum_C33_l = sin_phi_l * sunGeom->sindecl;

            }




            s0 = lumcline2(sunGeom, sunVarGeom, sunSlopeGeom, gridGeom,
                           horizonpointer);


            if (sunVarGeom->solarAltitude > 0.) {

                if ((!sunVarGeom->isShadow) && s0 > 0.) {

                    ra = bRadFunc(s0, &bh, sunVarGeom, sunSlopeGeom,
                                  &sunRadVar_cs);
                    beam_irr = ra;
                    if ((diff_rad != NULL) || (glob_pow != NULL)) {
                        dra =
                            dRadFunc(s0, bh, &rr, sunVarGeom, sunSlopeGeom,
                                     &sunRadVar_cs);
                        diff_irr = dra;
                    }
                    if ((refl_rad != NULL) || (glob_pow != NULL)) {
                        refl_irr = rr;
                    }
                    ra *= sunRadVar->cbh / sunRadVar_cs.cbh;
                    bh *= sunRadVar->cbh / sunRadVar_cs.cbh;
                    beam_irr_real = ra;
                    beam_e += dfr * ra;
                    ra = 0.;
                }
                else {
                    bh = 0.;
                    if ((diff_rad != NULL) || (glob_pow != NULL)) {
                        dra =
                            dRadFunc(s0, bh, &rr, sunVarGeom, sunSlopeGeom,
                                     &sunRadVar_cs);
                        diff_irr = dra;
                    }
                    if ((refl_rad != NULL) || (glob_pow != NULL)) {
                        refl_irr = rr;
                    }
                }
                if ((diff_rad != NULL) || (glob_pow != NULL)) {
                    dra =
                        dRadFunc(s0, bh, &rr, sunVarGeom, sunSlopeGeom,
                                 sunRadVar);
                    diff_e += dfr * dra;
                    diff_irr_real = dra;
                    dra = 0.;
                }
                if ((refl_rad != NULL) || (glob_pow != NULL)) {
                    if ((diff_rad == NULL) && (glob_pow == NULL))
                        dRadFunc(s0, bh, &rr, sunVarGeom, sunSlopeGeom,
                                 sunRadVar);
                    refl_e += dfr * rr;
                    refl_irr_real = rr;
                    rr = 0.;
                }
            }                   /* illuminated */

            totRad = beam_irr + diff_irr + refl_irr;
            totrad_real = beam_irr_real + diff_irr_real + refl_irr_real;


            if (useTemperatureCorr) {
                presTemperature =
                    temperatureInterpolate(temperatureData, presTime,
                                           longitude);
                effic = efficiency(totRad, presTemperature);
            }

            totRadCorr += effic * totrad_real * dfr;

            sunGeom->timeAngle = sunGeom->timeAngle + dfr_rad;

            if (sunGeom->timeAngle > lastAngle) {
                ss = 0;         /* we've got the sunset */
            }
            presTime += step;
        }                       /* end of while */
    }                           /* all-day radiation */


    return totRadCorr;
}

/*////////////////////////////////////////////////////////////////////// */


void where_is_point(double *length, struct SunGeometryVarDay *sunVarGeom,
                    struct GridGeometry *gridGeom)
{
    double sx, sy;

    double dx, dy;

    /*              double adx, ady; */
    int i, j;

    sx = gridGeom->xx0 * invstepx + offsetx;    /* offset 0.5 cell size to get the right cell i, j */
    sy = gridGeom->yy0 * invstepy + offsety;

    i = (int)sx;
    j = (int)sy;

    /*      if (i < n-1  && j < m-1) {    to include last row/col */
    if (i <= n - 1 && j <= m - 1) {

        dx = (double)i *gridGeom->stepx;

        dy = (double)j *gridGeom->stepy;

        *length = DISTANCE1(gridGeom->xg0, dx, gridGeom->yg0, dy);      /* dist from orig. grid point to the current grid point */

        sunVarGeom->zp = z[j][i];

        /*
           cube(j, i);
         */
    }
}

/*////////////////////////////////////////////////////////////////////// */

void calculate(int angleloss, double singleSlope, double singleAspect,
               double singleAlbedo, struct GridGeometry gridGeom)
{
    int i, j, l;

    /*                      double energy; */
    int someRadiation;

    int numRows;

    int arrayOffset;

    double latid_l, cos_u, cos_v, sin_u, sin_v;

    double sin_phi_l, tan_lam_l;

    double q1;

    double dayRad;

    double totrad;

    double sinAxisSlope, cosAxisSlope;

    double locTimeOffset;

    double latitude, longitude;

    struct SunGeometryConstDay sunGeom;

    struct SunGeometryVarDay sunVarGeom;

    struct SunGeometryVarSlope sunSlopeGeom;

    struct SolarRadVar sunRadVar;

    BeamRadFunc bRadFunc;

    DiffRadFunc dRadFunc;



    if (angleloss) {
        bRadFunc = brad_angle_loss;
        dRadFunc = drad_angle_loss;
    }
    else {
        bRadFunc = brad;
        dRadFunc = drad;
    }


    sunSlopeGeom.slope = singleSlope;
    sinAxisSlope = sin(singleSlope);
    cosAxisSlope = cos(singleSlope);
    sunSlopeGeom.aspect = singleAspect;
    sunRadVar.alb = singleAlbedo;
    /* In case of clear-sky calculation where there are no input beam and diffuse coefficients */
    sunRadVar.cbh = 1.0;
    sunRadVar.cdh = 1.0;

    sunGeom.sindecl = sin(declination);
    sunGeom.cosdecl = cos(declination);


    someRadiation = (beam_rad != NULL) || (diff_rad != NULL) ||
        (refl_rad != NULL) || (glob_pow != NULL);


    fprintf(stderr, "\n\n");


    if (beam_rad != NULL) {
        beam = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            beam[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                beam[j][i] = UNDEFZ;
        }
    }


    if (diff_rad != NULL) {
        diff = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            diff[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                diff[j][i] = UNDEFZ;
        }
    }

    if (refl_rad != NULL) {
        refl = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            refl[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                refl[j][i] = UNDEFZ;
        }
    }

    if (glob_pow != NULL) {
        globrad = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            globrad[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                globrad[j][i] = UNDEFZ;
        }
    }



    sunRadVar.G_norm_extra = com_sol_const(day);

    numRows = m / numPartitions;


    if (useCivilTime()) {
        /* We need to calculate the deviation of the local solar time from the 
           "local clock time". */
        dayRad = pi2 * day / 365.25;
        locTimeOffset =
            0.128 * sin(dayRad - 0.04887) + 0.165 * sin(2 * dayRad + 0.34383);

        /* Time offset due to timezone as input by user */

        locTimeOffset += civilTime;
        setTimeOffset(locTimeOffset);
    }
    else {
        setTimeOffset(0.);

    }




    for (j = 0; j < m; j++) {
        G_percent(j, m - 1, 2);

        if (j % (numRows) == 0) {
            INPUT_part(j);
            arrayOffset = 0;
            shadowoffset = 0;

        }
        sunVarGeom.zmax = zmax;


        for (i = 0; i < n; i++) {




            gridGeom.xg0 = gridGeom.xx0 = (double)i *gridGeom.stepx;

            gridGeom.yg0 = gridGeom.yy0 = (double)j *gridGeom.stepy;


            gridGeom.xp = xmin + gridGeom.xx0;
            gridGeom.yp = ymin + gridGeom.yy0;

            func = NULL;

            sunVarGeom.z_orig = z1 = sunVarGeom.zp = z[arrayOffset][i];

            if ((sunVarGeom.z_orig != UNDEFZ)) {
                if (aspin != NULL) {
                    o_orig = o[arrayOffset][i];
                    if (o[arrayOffset][i] != 0.) {
                        sunSlopeGeom.aspect = o[arrayOffset][i] * deg2rad;
                    }
                    else {
                        sunSlopeGeom.aspect = UNDEF;
                    }
                }
                if (slopein != NULL) {
                    sunSlopeGeom.slope = s[arrayOffset][i] * deg2rad;
                    sinAxisSlope = sin(sunSlopeGeom.slope);
                    cosAxisSlope = cos(sunSlopeGeom.slope);
                }
                if (linkein != NULL) {
                    sunRadVar.linke = li[arrayOffset][i];
                    li_max = AMAX1(li_max, sunRadVar.linke);
                    li_min = AMIN1(li_min, sunRadVar.linke);
                }
                if (albedo != NULL) {
                    sunRadVar.alb = a[arrayOffset][i];
                    al_max = AMAX1(al_max, sunRadVar.alb);
                    al_min = AMIN1(al_min, sunRadVar.alb);
                }
                if (latin != NULL) {
                    latitude = la[arrayOffset][i];
                    la_max = AMAX1(la_max, latitude);
                    la_min = AMIN1(la_min, latitude);
                    latitude = -latitude * deg2rad;
                }
                if ((G_projection() != PROJECTION_LL)) {

                    longitude = gridGeom.xp;
                    latitude = gridGeom.yp;

                    if (pj_do_proj(&longitude, &latitude, &iproj, &oproj) < 0) {
                        G_fatal_error("Error in pj_do_proj");
                    }

                    la_max = AMAX1(la_max, latitude);
                    la_min = AMIN1(la_min, latitude);
                }
                else {          /* ll projection */
                    longitude = gridGeom.xp;
                    latitude = gridGeom.yp;
                    la_max = AMAX1(la_max, latitude);
                    la_min = AMIN1(la_min, latitude);
                }
                if (useCivilTime()) {
                    longitTime = -longitude / 15.;
                }
                latitude *= deg2rad;
                longitude *= deg2rad;


                if (coefbh != NULL) {
                    sunRadVar.cbh = cbhr[arrayOffset][i];
                }
                if (coefdh != NULL) {
                    sunRadVar.cdh = cdhr[arrayOffset][i];
                }
                if (sunRadVar.cbh == UNDEFZ) {
                    /* Null value for beam or diffuse coefficient */
                    if (beam_rad != NULL)
                        beam[j][i] = (float)UNDEFZ;
                    if (diff_rad != NULL)
                        diff[j][i] = (float)UNDEFZ;
                    if (refl_rad != NULL)
                        refl[j][i] = (float)UNDEFZ;
                    if (glob_pow != NULL)
                        globrad[j][i] = UNDEFZ;
                    continue;
                }
                cos_u = cos(M_PI / 2 - sunSlopeGeom.slope);     /* = sin(slope) */
                sin_u = sin(M_PI / 2 - sunSlopeGeom.slope);     /* = cos(slope) */
                cos_v = cos(M_PI / 2 + sunSlopeGeom.aspect);
                sin_v = sin(M_PI / 2 + sunSlopeGeom.aspect);

                if (tt != NULL)
                    sunGeom.timeAngle = tim;

                gridGeom.sinlat = sin(-latitude);
                gridGeom.coslat = cos(-latitude);

                sin_phi_l =
                    -gridGeom.coslat * cos_u * sin_v +
                    gridGeom.sinlat * sin_u;
                latid_l = asin(sin_phi_l);

                q1 = gridGeom.sinlat * cos_u * sin_v +
                    gridGeom.coslat * sin_u;
                tan_lam_l = -cos_u * cos_v / q1;
                sunSlopeGeom.longit_l = atan(tan_lam_l);
                sunSlopeGeom.lum_C31_l = cos(latid_l) * sunGeom.cosdecl;
                sunSlopeGeom.lum_C33_l = sin_phi_l * sunGeom.sindecl;

                if (someRadiation) {
                    com_par_const(longitTime, &sunGeom, &gridGeom);
                    sr_min = AMIN1(sr_min, sunGeom.sunrise_time);
                    sr_max = AMAX1(sr_max, sunGeom.sunrise_time);
                    ss_min = AMIN1(ss_min, sunGeom.sunset_time);
                    ss_max = AMAX1(ss_max, sunGeom.sunset_time);
                    totrad =
                        joules2(&sunGeom, &sunVarGeom, &sunSlopeGeom,
                                &sunRadVar, &gridGeom,
                                tempdata +
                                numTemperatures() * (arrayOffset * n + i),
                                horizonarray + shadowoffset, latitude,
                                longitude, sinAxisSlope, cosAxisSlope,
                                bRadFunc, dRadFunc);
                    if (beam_rad != NULL)
                        beam[j][i] = (float)beam_e;
                    if (diff_rad != NULL)
                        diff[j][i] = (float)diff_e;
                    if (refl_rad != NULL)
                        refl[j][i] = (float)refl_e;
                    if (glob_pow != NULL)
                        globrad[j][i] = totrad;
                }

            }                   /* undefs */
            shadowoffset += arrayNumInt;
        }
        arrayOffset++;

    }
    fprintf(stderr, "\n");

    /* re-use &hist, but try all to initiate it for any case */
    /*   note this will result in incorrect map titles       */
    if (beam_rad != NULL) {
        Rast_write_history(beam_rad, &hist);
    }
    else if (diff_rad != NULL) {
        Rast_write_history(diff_rad, &hist);
    }
    else if (refl_rad != NULL) {
        Rast_write_history(refl_rad, &hist);
    }
    else if (glob_pow != NULL) {
        Rast_write_history(glob_pow, &hist);
    }
    else
        G_fatal_error
            ("Failed to init map history: no output maps requested!");


}


double com_declin(int no_of_day)
{
    double d1, decl;

    d1 = pi2 * no_of_day / 365.25;
    decl = asin(0.3978 * sin(d1 - 1.4 + 0.0355 * sin(d1 - 0.0489)));
    decl = -decl;
    /*      printf(" declination : %lf\n", decl); */

    return (decl);
}
